//************************* s609drv.c ***************************
//* This file contains the loadable Linux device driver for
//* Sensoray model 609 General Purpose IO access and
//* Crosspoint Switch control
//*
//* Autor:		Charlie X. Liu
//*
//* Revision:
//*			03/18/03   0.0   initial
//*
//* Copyright (C) 2003	Sensoray Co., Inc.
//**************************************************************
/*
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330,
Boston, MA 02111-1307, USA
****************************************************************/

/* Deal with CONFIG_MODVERSIONS */
//#include <linux/config.h>	/*retrieve the CONFIG_* macros */
#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS			/* force it on */
#endif
#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif

/* The necessary header files */
#include <linux/module.h>   /* Specifically, a module */
#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h>	/* for get_user and put_user */
#else
//#include <asm/segment.h>
#include <linux/mm.h>
#endif

#include <linux/kernel.h>   /* We're doing kernel work */
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/wrapper.h>
#include <linux/sched.h>
#include <linux/vmalloc.h>
#include <linux/ioport.h>
#include <linux/major.h>
#include <linux/pci.h>
#include <linux/ioctl.h>
#include <linux/errno.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>

#include "s609drv.h"

/* Usage counter */
static int use_count = 0;
/* global */
static WORD s609grabber = 0;	// only support one S609 grabber, now
static VSMODE vsmode[SYS_MAXGRABBERS][SYS_VOUT] = {{{0,0}}};

MODULE_DESCRIPTION("s609drv - Sensoray Model 609 driver module for General Purpose IO access and Crosspoint Switch control");
MODULE_AUTHOR("Charlie X. Liu at Sensoray Co.");
MODULE_LICENSE("GPL");

// function prototypes
static int  s609_open ( struct inode *inode, struct file *file );
#if LINUX_VERSION_CODE>KERNEL_VERSION(2,2,0)
static int  s609_release ( struct inode *inode, struct file *file );
#else
static void s609_release ( struct inode *inode, struct file *file );
#endif
static int  s609_ioctl(struct inode *inode, struct file *file,
			     unsigned int cmd, unsigned long arg );
int  init_module ( void );
void cleanup_module ( void );
//
unsigned long RegRead ( unsigned long reg_addr );
void RegWriteDirect ( unsigned long reg_addr, unsigned long data );
void SetSwitch ( WORD output );
void InitSwitch ( void );
void LoadSerial ( DWORD * cmdword );
void LoadTermination ( BYTE term);


/* The access methods are placed in a table of valid file operations. This
   driver supports open, release (close), read, write and ioctl.
   Use the new gcc extension syntax.									*/
static struct file_operations s609_fops=
    {
	//read:		s609_read,	// read
	//write:	s609_write,	// write
	ioctl:		s609_ioctl,	// ioctl
	open:		s609_open,	// open
	release:	s609_release	// release
    };


//*************************************************************************
//* Open method - Enables the s609
//*

static int s609_open (struct inode *inode, struct file *file)
{
	if(MOD_IN_USE)        // already open (we use exclusive access)
	{
	#ifdef DEBUG
		printk("s609: Device already open\n");
	#endif
		return(-EBUSY);
	}

	use_count++;
	MOD_INC_USE_COUNT;

	return 0;

}


//**************************************************************************
//* Close method - Disables the s609
//*

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static int  s609_release(struct inode *inode, struct file *file)
#else
static void s609_release(struct inode *inode, struct file *file)
#endif
{
	use_count--;
	MOD_DEC_USE_COUNT;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	return 0;
#endif
}


//*****************************************************************************
//* Register and init module:
//*

static struct pci_dev       *dev = NULL;
static struct pci_device_id s609_pci_tbl[] = {
				{ F878_VENDOR_ID, F878_DEVICE_ID,
				  S609_SUBSYS_VENDOR_ID, S609_SUBSYS_ID,
				  0, 0, 0 },
				{0,}
};
static u32    s609_base_address;

//* This is called when the module is loaded.
int init_module (void)
{
	int result;
	// for signature checking
	void *cookie_virt;
	unsigned long kdat = 0;

	// start to install driver for Sensoray Model 609
	printk ("\ns609drv: Installing s609 device driver...\n");
	
	// finding Sensoray 609 in the system
	dev = pci_find_device(s609_pci_tbl[0].vendor, s609_pci_tbl[0].device, dev);
	if ( dev==NULL ) {
		printk(KERN_INFO "s609drv: there is no Fusion878 based Frame Grabber Board in the system.\n");
		unregister_chrdev(S609_MAJOR, "s609drv");
		return -1;
	}
	if ( dev->subsystem_vendor!=S609_SUBSYS_VENDOR_ID
					|| dev->subsystem_device!= S609_SUBSYS_ID ) {
		printk(KERN_INFO "s609drv: there is no Sensoray Model 609 in the system.\n");
		unregister_chrdev(S609_MAJOR, "s609drv");
		return -1;
	}
	printk ("s609drv: Sensoray Model 609 has been found.\n");
	// and
	pci_read_config_dword(dev, PCI_BASE_ADDRESS_0, &s609_base_address);
	s609_base_address &= 0xfffff000;	//since [11:0] = 0x008
	//printk ("s609drv: 	s609_base_address = 0x%x \n", s609_base_address );

	/* Register the device */
	if ((result=register_chrdev(S609_MAJOR, "s609drv", &s609_fops)))
	{
		printk ("Unable to get major number %i for s609.\n", S609_MAJOR);
		printk ("s609 device driver installation failed!\n");
		return result;
	}

	use_count = 0;

	//signature checking
	//cookie_virt = ioremap(Bt878_BASE_ADDR, 0x1000);
	cookie_virt = ioremap(s609_base_address, 0x1000);
	if (!cookie_virt)
		printk(KERN_ERR "init_module: unable to map region\n");
	else {
		kdat = readl(cookie_virt+GPIO_DATA);
		rmb();
		iounmap(cookie_virt);
	}
	if ((kdat&0x800000)!=0) {
		printk ("s609drv: S609 device driver installation failed!\n");
		printk ("     the s609drv is specifically designed for Sensoray Model 609.\n");
		unregister_chrdev(S609_MAJOR, "s609drv");
		return -1;
	}
	// initialize crosspoint switch
	InitSwitch();

	//successful
	printk ("s609drv: S609 device driver successfuly installed!\n");
	return 0;
}


//*****************************************************************************
//* Unregister device module and cleanup:
//*					  This is called when the module is unloaded.

void cleanup_module (void)
{

	/* unregister device */
	if (unregister_chrdev(S609_MAJOR, "s609drv")<0)
		printk ("s609drv: S609 device driver: Module unloading error\n");
	else
		printk ("s609drv: S609 device driver removed!\n");

}


//*****************************************************************************
//* ioctl function
//*
//* Args:	inode	- inode structure
//*			file	- file structure
//*			cmd		- command to execute
//*			arg		- argument, is used as pointer to ioparam
//*

static int s609_ioctl(struct inode *inode, struct file *file,
			 unsigned int cmd, unsigned long arg)
{
  struct ioparam request;

    //printk("cmd = %X \n", cmd);
    //printk("RegRD: = %X \n", RegRD);
    //printk("RegWR: = %X \n", RegWR);

  switch(cmd)
  {
	case  RegRD:		// read from reg :  dat <- reg_addr
	{
	    unsigned long kdat = 0;
	    unsigned long reg_addr;
	    void *cookie_virt;

	    copy_from_user(&request, (struct ioparam *) arg, sizeof(request));
	    //reg_addr = Bt878_BASE_ADDR + request.param2;	// reg addr
	    reg_addr = s609_base_address + request.param2;	// reg addr

	    //if (reg_addr<Bt878_BASE_ADDR||reg_addr>(Bt878_BASE_ADDR+0x1000))
	    if ( reg_addr<s609_base_address || reg_addr>(s609_base_address+0x1000) )
			return(-EINVAL);
	    //printk("Read: reg = %X \n", (int)reg_addr);		// for debugging
		//instead,
		//cookie_virt = ioremap(Bt878_BASE_ADDR, 0x1000);
		cookie_virt = ioremap(s609_base_address, 0x1000);
		if (!cookie_virt)
			printk(KERN_ERR "Read: unable to map region\n");
		else {
			//kdat = readl(cookie_virt+(reg_addr-Bt878_BASE_ADDR));
			kdat = readl(cookie_virt+(reg_addr-s609_base_address));
			rmb();
			iounmap(cookie_virt);
		}

	    //printk("Read: data=%X <- reg:%X\n", (int)kdat, (int)reg_addr-Bt878_BASE_ADDR);
	    request.param1 = kdat;
	    copy_to_user((struct ioparam *) arg, &request, sizeof(request));

	    break;
	}

	case  RegWR:		// write to reg :  dat -> reg_addr
	{
	    unsigned long dat; unsigned long reg_addr;
	    void *cookie_virt;

	    copy_from_user(&request, (struct ioparam *) arg, sizeof(request));
	    dat = request.param1;	//data
	    //reg_addr = Bt878_BASE_ADDR + request.param2;	// reg addr
	    reg_addr = s609_base_address + request.param2;	// reg addr
	    //printk("Write: reg = %X \n", reg_addr);		// for debugging

	    //if (reg_addr<Bt878_BASE_ADDR||reg_addr>(Bt878_BASE_ADDR+0x1000))
	    if ( reg_addr<s609_base_address || reg_addr>(s609_base_address+0x1000) )
		return(-EINVAL);
	    //printk("Write: data=%X -> reg:%X\n", (int)dat, (int)reg_addr-Bt878_BASE_ADDR);
		//instead,
		//cookie_virt = ioremap(Bt878_BASE_ADDR, 0x1000);
		cookie_virt = ioremap(s609_base_address, 0x1000);
		if (!cookie_virt)
			printk(KERN_ERR "Write: unable to map bt878_region\n");
		else {
			//writel(dat, (cookie_virt+(reg_addr-Bt878_BASE_ADDR)));
			writel(dat, (cookie_virt+(reg_addr-s609_base_address)));
			wmb();
			iounmap(cookie_virt);
		}

	    break;
	}

	case  DioRD:		// read from digital IO :  dat
	{
	    unsigned long kdat = 0;
	    DWORD         data;

	    //set data bits of GPIO to read mode
	    RegWriteDirect ( GPIO_OUT_EN, GPIO_EN_RD);
	    //enable read buffer
	    data = CTRL_IDLE & ~IO_RD;
	    RegWriteDirect (GPIO_DATA, data);
	    kdat = (WORD) RegRead (GPIO_DATA);

	    //printk("Read: data=%X <- reg:%X\n", (int)kdat, (int)reg_addr);
	    request.param1 = kdat;
	    copy_to_user((struct ioparam *) arg, &request, sizeof(request));

	    break;
	}

	case  DioWR:		// write to digital IO :  dat
	{
	    DWORD data, iodata;

	    copy_from_user(&request, (struct ioparam *) arg, sizeof(request));
	    iodata = request.param1;	//data

	    //set data bits of GPIO to write mode
	    RegWriteDirect ( GPIO_OUT_EN, GPIO_EN_WR);
	    //data setup
	    data = CTRL_IDLE | iodata;
	    RegWriteDirect ( GPIO_DATA, data);
	    //clock
	    RegWriteDirect ( GPIO_DATA, data | IO_WR);
	    RegWriteDirect ( GPIO_DATA, CTRL_IDLE);
	    //set data bits of GPIO back to read mode
	    RegWriteDirect ( GPIO_OUT_EN, GPIO_EN_RD);

	    break;
	}

	case  XptSW:		// CrossPoint Switch :  input# (0-15) -> output# (0-7)
	{
	    WORD input, output;
	    static BOOL term = TRUE;	// always TRUE, for now

	    copy_from_user(&request, (struct ioparam *) arg, sizeof(request));
	    input  = request.param1;		//video input channel number (1-16) <-- (1-16)
	    output = request.param2 - 1;	//video output channel number (0-7) <--  (1-8)

	    if ( input<=0 || input>SYS_VIN ) {			// valid in [1..SYS_VIN]
	        printk("XptSW: invalid video input channel # %d, should be in (1-16) \n", input+1);
		return(-EINVAL);
	    }
	    if ( /* output<0 || */ output>=SYS_VOUT ) {		// valid out [0..(SYS_VOUT-1)]

	        printk("XptSW: invalid video output channel # %d, should be in (1-8) \n", output+1);
		return(-EINVAL);
	    }

	    //set system status
	    vsmode[s609grabber][output].input = input;
	    vsmode[s609grabber][output].termination = term;

	    SetSwitch (output);

	    break;
	}

	default: return(-EINVAL);   // invalid function number

  }

  return(0);

}


//*---------------------------------------------------------------
//* sub functions
//*---------------------------------------------------------------

//  basic func
unsigned long RegRead ( unsigned long reg_addr )
{
	unsigned long data = 0;

	void *cookie_virt;

	if ( reg_addr<0 || reg_addr>0x1000)
		return(-EINVAL);
	//cookie_virt = ioremap(Bt878_BASE_ADDR, 0x1000);
	cookie_virt = ioremap(s609_base_address, 0x1000);
	if (!cookie_virt)
		printk(KERN_ERR "RegRead: unable to map region\n");
	else {
		data = readl(cookie_virt+reg_addr);
		rmb();
		iounmap(cookie_virt);
	}
	//printk("RegRead: data=%X <- reg:%X\n", (int)data, (int)reg_addr);

	return data;

}

void RegWriteDirect ( unsigned long reg_addr, unsigned long data )
{
	void *cookie_virt;

	if ( reg_addr<0 || reg_addr>0x1000 )
	{
		printk(KERN_ERR "RegWrite: invalid reg_addr!\n");
		return;
	}

	//printk("RegWrite: data=%X -> reg:%X\n", (int)data, (int)reg_addr);
	//cookie_virt = ioremap(Bt878_BASE_ADDR, 0x1000);
	cookie_virt = ioremap(s609_base_address, 0x1000);
	if (!cookie_virt)
		printk(KERN_ERR "RegWrite: unable to map bt878_region\n");
	else {
		writel(data, (cookie_virt+reg_addr));
		wmb();
		iounmap(cookie_virt);
	}

	return;
}

//  switching
void SetSwitch ( WORD output )
{
	DWORD cwindex;
	BYTE i, shifts, xinput, term = 0;
	DWORD cword[2] = {0xaaaaaaaa, 0xaaaaaaaa};			//turn of all
	BOOL upper;							//TRUE if input channel [9..16], cword[0]
	DWORD ormask, andmask;

	for (i = 0; i < SYS_VOUT; i++) {				//for all 8 outputs
		shifts = (7 - i) * 4;					//shifts # to get to the nibble corresp. to current output
		andmask = ~(0xf << shifts);				//to clear corresponding nibble
		if (vsmode[s609grabber][i].input > 0) {			//if any input selected for current output
			upper = (vsmode[s609grabber][i].input > 8);
			cwindex = (upper ? 0 : 1);			//which cword[] are we modifying
			xinput = vsmode[s609grabber][i].input - (upper ? 9 : 1);
			if (i == output) {
				ormask = (SYS_ACTIVATE << shifts);	//to activate channel
			} else {
				ormask = (xinput << shifts);
			}
			cword[cwindex] = (cword[cwindex] & andmask) | ormask;
		}
	}
	LoadSerial ( cword );

	for (i = 0; i < SYS_VOUT; i++) {				//for all 8 outputs
		shifts = (7 - i) * 4;					//shifts # to get to the nibble corresp. to current output
		andmask = ~(0xf << shifts);				//to clear corresponding nibble
		if (vsmode[s609grabber][i].input > 0) {			//if any input selected for current output
			upper = (vsmode[s609grabber][i].input > 8);
			cwindex = (upper ? 0 : 1);			//which cword[] are we modifying
			xinput = vsmode[s609grabber][i].input - (upper ? 9 : 1);
			ormask = (xinput << shifts);
			cword[cwindex] = (cword[cwindex] & andmask) | ormask;
		}
		term |= ((vsmode[s609grabber][i].termination ? 0 : 1) << i);	//set term bit to 0 if ON (max4677)
	}
	LoadSerial ( cword );

	LoadTermination ( term );

	return;
}

void InitSwitch ( void )
{
	DWORD cword[2] = {0x99999999, 0x99999999};

	LoadSerial (cword);
	return;
}

void LoadSerial ( DWORD * cmdword )
{
	int i, j;
	DWORD data, mask;

	RegWriteDirect ( GPIO_DATA, CTRL_IDLE);			//reset ctrl bits to defaults

	//Set CE# to 0
	data = CTRL_IDLE & ~SER_CE;
	RegWriteDirect ( GPIO_DATA, data);

	for (j = 0; j < 2; j++) {
		for (i = 31; i >= 0; i--) {			//bit by bit starting with MSB
			mask = 1 << i;				//current bit mask
			//set data bit in control word equal to current bit in cmdword
			data = (CTRL_IDLE & ~SER_DATA & ~SER_CE) | ((cmdword[j] & mask) ? SER_DATA : 0);
			//write same data several times to provide required delays
			//(each write is at least 33 ns, possibly more)
			RegWriteDirect ( GPIO_DATA, data);
			RegWriteDirect ( GPIO_DATA, data);
			RegWriteDirect ( GPIO_DATA, data);
			RegWriteDirect ( GPIO_DATA, data);
			//set WR bit to 1
			data |= SER_WR;
			RegWriteDirect ( GPIO_DATA, data);
			RegWriteDirect ( GPIO_DATA, data);
			RegWriteDirect ( GPIO_DATA, data);
			//set WR bit to 0
			data &= ~SER_WR;
			RegWriteDirect ( GPIO_DATA, data);
			RegWriteDirect ( GPIO_DATA, data);
			RegWriteDirect ( GPIO_DATA, data);
		}
	}

	//Latch the new data (~|_|~)
	data &= ~SER_LATCH;
	RegWriteDirect ( GPIO_DATA, data);
	RegWriteDirect ( GPIO_DATA, data);
	RegWriteDirect ( GPIO_DATA, data);
	data |= SER_LATCH;
	RegWriteDirect ( GPIO_DATA, data);
	RegWriteDirect ( GPIO_DATA, data);
	RegWriteDirect ( GPIO_DATA, data);


	//Set CE# to 1
	RegWriteDirect ( GPIO_DATA, CTRL_IDLE);

	return;
}

void LoadTermination ( BYTE term)
{
	DWORD data;

	//set data bits of GPIO to write mode
	RegWriteDirect ( GPIO_OUT_EN, GPIO_EN_WR);
	data = CTRL_IDLE | term;
	//data setup
	RegWriteDirect ( GPIO_DATA, data);
	//clock
	RegWriteDirect ( GPIO_DATA, data | TERM_CLK);
	RegWriteDirect ( GPIO_DATA, CTRL_IDLE);
	//set data bits of GPIO back to read mode
	RegWriteDirect ( GPIO_OUT_EN, GPIO_EN_RD);

	return;
}

